# import uuid
# from filelock import FileLock
#
# from b_logger import tmp_logs_path
# from b_logger.utils.basedatamodel import BaseDataModel
# from b_logger.entities.statuses import TestStatus
# from b_logger.entities.tests import TestReport
#
#
# class ModuleTestResults(BaseDataModel):
#     def __init__(self):
#         self.passed: int = 0
#         self.failed: int = 0
#         self.broken: int = 0
#         self.skipped: int = 0
#
#     def increase(self, result: TestStatus):
#         if not hasattr(self, result):
#             raise ValueError(f'Unsupported test status: {result}')
#         current = getattr(self, result)
#         setattr(self, result, current + 1)
#
#
# class ModuleTestReports(BaseDataModel, list):
#     def __init__(self):
#         super().__init__()
#
#     def add_test_report(self, test_report: TestReport):
#         self.append(test_report)
#
#     def get_test_report_by_id(self, test_rep_id):
#         for rep in self:
#             if rep.id == test_rep_id:
#                 return rep
#             else:
#                 raise ValueError(f'No such test run_report id: {test_rep_id}')
#
#
# class ModuleData(BaseDataModel):
#     def __init__(self, module_name):
#         self.module_id = f'modules_{uuid.uuid4()}'
#         self.module_name = module_name
#         self.module_results = ModuleTestResults()
#         self.module_tests = ModuleTestReports()
#
#     def add_module_result(self, result: TestStatus):
#         self.module_results.increase(result)
#
#     def add_test_report(self, test_report: TestReport):
#         test_name = test_report.name
#         self.module_tests[test_name] = test_report
#
#     def save_json(self, file_name=None):
#         root = f'{tmp_logs_path()}/modules'
#         if file_name:
#             path = f'{root}/{file_name}'
#         else:
#             path = f'{root}/{self.module_id}'
#         with FileLock(f'{path}.lock'):
#             self.to_json_file(path)
#
#
# class ModuleDataContainer(BaseDataModel, dict):
#     def __init__(self, module_name):
#         super().__init__()
#
#     def add_module(self, module_data: ModuleData):
#         self[module_data.module_name] = module_data.module_id
#


import os
import json
import uuid
from collections import defaultdict
from datetime import datetime
from dateutil import parser
from filelock import FileLock

from b_logger.config import blog_config
from b_logger.entities.statuses import TestStatus
from b_logger.entities.tests import TestReport
from b_logger.entities.steps import StepContainer
from b_logger.utils.basedatamodel import BaseDataModel
from b_logger.utils.paths import b_logs_path, b_logs_tmp_path, b_logs_tmp_steps_path, b_logs_tmp_reports_path


class RunResults(BaseDataModel):
    def __init__(self):
        self.PASSED: int = 0
        self.FAILED: int = 0
        self.BROKEN: int = 0
        self.SKIPPED: int = 0
        self.NONE: int = 0

    def increase(self, result: TestStatus, amount: int = 1):
        if not hasattr(self, result):
            raise ValueError(f'Unsupported test status: {result}')
        current = getattr(self, result)
        setattr(self, result, current + amount)


class RunReport(BaseDataModel):
    def __init__(self):
        self.report_id = f'report_{uuid.uuid4()}'
        self.proj_name = blog_config.project_name
        self.base_url = blog_config.base_url
        self.env = blog_config.env
        self.worker = None
        self.start_time = datetime.now().strftime('%d/%m/%Y %H:%M:%S')
        self.end_time = None
        self.duration = None
        self.report_ids = {}
        self.run_results = RunResults()
        self.modules: dict[str, dict] = defaultdict(
            lambda: {
                "results": RunResults(),
                "tests": defaultdict(list)
            }
        )

    def set_base_url(self, base_url: str):
        self.base_url = base_url

    def set_env(self, env: str):
        self.env = env

    def set_worker(self, worker):
        self.worker = worker

    def set_end_time(self):
        self.end_time = datetime.now().strftime('%d/%m/%Y %H:%M:%S')

    def count_duration(self):
        if isinstance(self.start_time, str) or isinstance(self.end_time, str):
            self.start_time = self.get_iso_start_time()
            self.end_time = self.get_iso_end_time()

        self.duration = self.end_time - self.start_time

    def get_iso_start_time(self):
        if isinstance(self.start_time, str):
            self.start_time = parser.parse(self.start_time)
        return self.start_time

    def get_iso_end_time(self):
        if isinstance(self.end_time, str):
            self.end_time = parser.parse(self.end_time)
        return self.end_time

    def add_result(self, test_report: TestReport):
        module = test_report.module
        test_name = test_report.originalname
        status = test_report.status

        self.modules[module]['tests'][test_name].append(test_report)

        self.modules[module]['results'].increase(status)
        self.run_results.increase(status)

    def get_steps(self) -> dict:
        steps_by_id = {}
        if self.modules:
            for module_data in self.modules.values():
                for test_name, test_reports in module_data['tests'].items():
                    for report in test_reports:
                        steps_id = report.get('steps')
                        if steps_id:
                            path = f'{b_logs_tmp_steps_path()}/{steps_id}.json'
                            try:
                                steps_by_id[steps_id] = StepContainer.from_json(path)
                            except FileNotFoundError:
                                continue
        return steps_by_id

    def combine_modules_from_report(self, run_report):
        if run_report.modules:
            for module_name, module_data in run_report.modules.items():
                mod_results = module_data["results"]
                mod_tests: dict = module_data["tests"]

                module = self.modules[module_name]

                for status, count in mod_results.items():
                    module['results'].increase(status, count)

                for test_name, test_reports in mod_tests.items():
                    module['tests'][test_name].extend(test_reports)

    def save_json(self, file_name=None):
        root = f'{b_logs_tmp_reports_path()}'
        if file_name:
            path = f'{root}/{file_name}'
        else:
            path = f'{root}/{self.report_id}'
        with FileLock(f'{path}.lock'):
            self.to_json_file(path)

    # def get_modules(self):
    #     return self.modules.keys()



import re
import shutil
import uuid
import mimetypes
import json
import os
from pathlib import Path
from typing import Union, Optional, BinaryIO

from b_logger.utils.basedatamodel import BaseDataModel
from b_logger.utils.paths import attachments_path


class Attachment(BaseDataModel):
    root = Path(attachments_path())

    def __init__(
        self,
        content: Union[bytes, Path, BinaryIO, str, dict, list, int, float, bool, None] = None,
        name: Optional[str] = None,
        type_: Optional[str] = None,
        _skip_processing: bool = False,
    ):
        self.name = self.normalize_name(name) if name else f'attachment_{uuid.uuid4()}'
        self.type_ = type_

        if not _skip_processing:
            self._process_attachment(content)

    def _process_attachment(self, content):
        # 1. bytes
        if isinstance(content, bytes):
            self.type_ = 'image/png'
            self.ensure_extension('.png')
            self.__from_bytes(content)

        # 2. JSON supported
        elif isinstance(content, (dict, list, int, float, bool, type(None))):
            self.type_ = 'application/json'
            self.ensure_extension('.json')
            data = json.dumps(content, ensure_ascii=False, indent=2).encode('utf-8')
            self.__from_bytes(data)

        # 3. str
        elif isinstance(content, str):
            if not content.strip():
                print('[BLogger][WARN] Cannot attach empty string')

            # path = Path(content)
            # if path.exists() and path.is_file():
            #     self.type_ = mimetypes.guess_type(str(path))[0] or 'application/octet-stream'
            #     self.__from_file(path)
            # else:

            try:
                parsed = json.loads(content)
                self.type_ = 'application/json'
                self.ensure_extension('.json')
                formatted = json.dumps(parsed, ensure_ascii=False, indent=2).encode('utf-8')
                self.__from_bytes(formatted)
            except json.JSONDecodeError:
                self.type_ = 'text/plain'
                self.ensure_extension('.txt')
                self.__from_bytes(content.encode('utf-8'))

        # 4. Path
        elif isinstance(content, Path):
            if not content.exists() or not content.is_file():
                print(f'[BLogger][WARN] Attachment path is invalid: {content}')
            self.type_ = mimetypes.guess_type(str(content))[0] or 'application/octet-stream'
            self.__from_file(content)

        # 5. File-like (BinaryIO)
        elif hasattr(content, 'read'):
            self.type_ = mimetypes.guess_type(getattr(content, 'name', ''))[0] or 'application/octet-stream'
            self.__from_filelike(content)

        else:
            print(f'[BLogger][WARN] Unsupported attachment content type: {type(content)}')

    def __from_file(self, file_path: Path):
        if not file_path.exists():
            raise FileNotFoundError(f"Attachment file not found: {file_path}")
        self.ensure_extension(file_path.suffix or '')
        dest_path = self.root / self.name
        if not dest_path.exists():
            shutil.copy2(file_path, dest_path)

    def __from_bytes(self, data: bytes):
        dest_path = self.root / self.name
        with open(dest_path, 'wb') as f:
            f.write(data)

    def __from_filelike(self, file_obj: BinaryIO):
        data = file_obj.read()
        if isinstance(data, str):
            data = data.encode('utf-8')
        self.__from_bytes(data)

    @classmethod
    def from_dict(cls, data: dict) -> 'Attachment':
        name = data.get("name")
        type_ = data.get("type_")
        path = cls.root / name
        if not path.exists():
            raise FileNotFoundError(f'[BLogger][WARN] Attachment file not found: {path}')
        return cls(path, name=name, type_=type_, _skip_processing=True)

    def ensure_extension(self, ext: str):
        if not self.name.endswith(ext):
            self.name += ext

    @staticmethod
    def normalize_name(name: str):
        name = re.sub(r'[^A-Za-z0-9_-]', '_', name)
        name = re.sub(r'_+', '_', name).strip('_')

        existing_names = {Path(att).stem for att in os.listdir(f'{attachments_path()}')}
        base_name = name
        index = 0
        while name in existing_names:
            index += 1
            name = f'{base_name}_{index}'

        return name


# class Attachment(BaseDataModel):
#     root = Path(attachments_path())
#
#     def __init__(
#         self,
#         content: Union[bytes, Path, BinaryIO, str, dict, list, int, float, bool, None] = None,
#         name: Optional[str] = None,
#         type_: Optional[str] = None,
#         _skip_processing: bool = False,
#     ):
#         self.name = name if name else f'attachment_{uuid.uuid4()}'
#         self.type_ = type_
#
#         if not _skip_processing:
#             self._process_attachment(content)
#
#     def _process_attachment(self, content):
#         # 1. bytes
#         if isinstance(content, bytes):
#             self.type_ = mimetypes.types_map.get('.png', 'image/png')
#             self.ensure_extension('.png')
#             self.__from_bytes(content)
#
#         # 2. JSON-serializable (dict, list, int, float, bool, None)
#         elif isinstance(content, (dict, list, int, float, bool, type(None))):
#             self.type_ = mimetypes.types_map.get('.json', 'application/json')
#             self.ensure_extension('.json')
#             data = json.dumps(content, ensure_ascii=False, indent=2).encode('utf-8')
#             self.__from_bytes(data)
#
#         # 3. str
#         elif isinstance(content, str):
#             if not content.strip():
#                 raise ValueError("[BLogger][WARN] Cannot attach empty string")
#
#             try:
#                 parsed = json.loads(content)
#                 self.type_ = mimetypes.types_map.get('.json', 'application/json')
#                 self.ensure_extension('.json')
#                 formatted = json.dumps(parsed, ensure_ascii=False, indent=2).encode('utf-8')
#                 self.__from_bytes(formatted)
#             except json.JSONDecodeError:
#                 self.type_ = mimetypes.types_map.get('.txt', 'text/plain')
#                 self.ensure_extension('.txt')
#                 self.__from_bytes(content.encode('utf-8'))
#
#         # 4. Path
#         elif isinstance(content, Path):
#             if not content.exists() or not content.is_file():
#                 raise FileNotFoundError(f'[BLogger][WARN] Attachment path is invalid: {content}')
#             self.type_ = mimetypes.guess_type(str(content))[0] or 'application/octet-stream'
#             self.ensure_extension(content.suffix)
#             self.__from_file(content)
#
#         # 5. File-like object
#         elif hasattr(content, 'read'):
#             self.type_ = mimetypes.guess_type(getattr(content, 'name', ''))[0] or 'application/octet-stream'
#             self.ensure_extension(Path(getattr(content, 'name', '')).suffix or '.bin')
#             self.__from_filelike(content)
#
#         else:
#             raise TypeError(f"[BLogger][WARN] Unsupported attachment content type: {type(content)}")
#
#     def __from_file(self, file_path: Path):
#         if not file_path.exists():
#             raise FileNotFoundError(f"Attachment file not found: {file_path}")
#         dest_path = self.root / self.name
#         if not dest_path.exists():
#             shutil.copy2(file_path, dest_path)
#
#     def __from_bytes(self, data: bytes):
#         dest_path = self.root / self.name
#         dest_path.parent.mkdir(parents=True, exist_ok=True)
#         with open(dest_path, 'wb') as f:
#             f.write(data)
#
#     def __from_filelike(self, file_obj: BinaryIO):
#         if isinstance(file_obj, TextIOBase):
#             data = file_obj.read().encode('utf-8')
#         else:
#             data = file_obj.read()
#         self.__from_bytes(data)
#
#     def ensure_extension(self, ext: str):
#         ext = ext.lower()
#         if not ext.startswith('.'):
#             ext = f'.{ext}'
#         current_ext = Path(self.name).suffix.lower()
#         if current_ext != ext:
#             base = Path(self.name).stem
#             self.name = f"{base}{ext}"
#
#     @classmethod
#     def from_dict(cls, data: dict) -> 'Attachment':
#         name = data.get("name")
#         type_ = data.get("type_")
#         if not name:
#             raise ValueError("Missing 'name' field in attachment dictionary")
#         path = cls.root / name
#         if not path.exists():
#             raise FileNotFoundError(f'[BLogger][WARN] Attachment file not found: {path}')
#         return cls(path, name=name, type_=type_, _skip_processing=True)
#
#     def to_dict(self) -> dict:
#         return {
#             "name": self.name,
#             "type_": self.type_,
#         }
#
#     def __repr__(self):
#         return f"<Attachment name={self.name!r}, type={self.type_!r}>"



import json
import traceback
from datetime import datetime
from pathlib import Path
from typing import Union, BinaryIO

from b_logger.entities.reports import RunReport
from b_logger.entities.tests import TestReport, TestStatus
from b_logger.entities.attachments import Attachment
from b_logger.entities.prints import Print
from b_logger.entities.steps import Step, StepStatus, StepError, StepContainer
from b_logger.entities.statuses import py_outcome_to_tstatus
from b_logger.integrations import Integrations
from b_logger.utils.browser_adapters import get_browser_adapter
from b_logger.utils.formatters import format_tb
from b_logger.utils.paths import attachments_path


class TestRun:
    pass


class RunTime:
    def __init__(self):
        self.run_report: RunReport = RunReport()
        self.browser: "RemoteWebDriver | WebDriver | Page | None" = None
        self.test_report: TestReport = TestReport()
        self.step_container: StepContainer = StepContainer()

    def set_base_url(self, base_url: str):
        self.run_report.set_base_url(base_url)

    def set_env(self, env: str):
        self.run_report.set_env(env)

    def set_browser(self, browser: "RemoteWebDriver | WebDriver | Page"):
        self.browser = browser

    def start_test(self, item):
        module = item.location[0]
        test_name = item.name
        test_originalname = item.originalname

        self.step_container = StepContainer()
        self.test_report = TestReport(module, test_name, test_originalname)

    def finish_test(self):
        self.step_container.save_json()
        self.test_report.set_steps(self.step_container.container_id)

        self.run_report.add_result(self.test_report)

        del self.step_container, self.test_report
        if self.browser:
            self.browser = None

    def process_test_result(self, report, call, item):
        """Process test results and set appropriate status."""

        self.test_report.set_duration(round(report.duration, 2))

        if report.longrepr:
            self.test_report.set_stacktrace(report.longreprtext)

        if report.failed:
            self._handle_failed_test(call, report, item)

        elif report.skipped:
            self._handle_skipped_test(call, report, item)

        else:
            self._handle_passed_test(call, report, item)

    def _handle_failed_test(self, call, report, item):
        self.make_screenshot(is_error=True)
        self.test_report.set_error(call.excinfo.exconly())

    def _handle_skipped_test(self, call, report, item):
        self.make_screenshot(is_error=True)
        self.test_report.set_error(call.excinfo.exconly())

    def _handle_passed_test(self, call, report, item):
        pass

    def process_test_status(self, report, call, item):
        if hasattr(report, 'wasxfail'):
            if report.outcome == 'skipped':
                status = TestStatus.PASSED
            else:
                status = TestStatus.FAILED
        else:
            if call.excinfo and report.outcome == 'failed':
                if call.excinfo.typename == 'AssertionError':
                    status = TestStatus.FAILED
                else:
                    status = TestStatus.BROKEN
            else:
                status = py_outcome_to_tstatus(report.outcome)

        self.test_report.set_status(status)

        return status

    def start_step(self, step: Step):
        if self.step_container.current_step_id is not None:
            parent = self.step_container.get_current_step()
            step.set_parent_id(parent.id)

            parent.add_sub_step(step)

            self.step_container.set_current_step(step.id)

        else:
            self.step_container.set_current_step(step.id)
            self.step_container.add_step(step)

    def handle_step_result(self, step: Step, exc=None):
        if exc:
            if not self.step_container.failed:
                self.make_step_err_scr(step)

                step.set_error(StepError(exc, format_tb(traceback.format_exc(4))))

                self.step_container.failed = True

            if step.parent_id is None:
                self.step_container.failed = False

            step.set_status(StepStatus.FAILED)

            return

        step.set_status(StepStatus.PASSED)

    def finish_step(self, step: Step):
        self.step_container.set_current_step(step.parent_id)

    def apply_description(self, description):
        if not self.test_report.description:
            self.test_report.set_description(description)
            Integrations.description(description)
        else:
            self.test_report.modify_description(description)

    def apply_info(self, **kwargs):

        if not kwargs:
            print('[BLogger][WARN] blog.info() requires at least one keyword argument')
            return

        info = {}

        for k, v in kwargs.items():
            key = k.replace('_', ' ').capitalize().upper()

            info[key] = v

            Integrations.info(k, v)

        current_step = self.step_container.get_current_step()

        if current_step:
            current_step.add_info(info)

        self.test_report.add_info(info)

    def apply_link(self, **kwargs):

        links = {}

        for k, v in kwargs.items():
            key = k.replace('_', ' ').capitalize().upper()

            links[key] = f'<a href={v}>{v}</a>'

            Integrations.link(v, key)

        self.apply_info(links=links)

        # links = {}
        #
        # for k, v in kwargs.items():
        #     key = k.replace('_', ' ').capitalize().upper()
        #
        #     links[key] = f'<a href={v}>{v}</a>'
        #
        # current_step = self.step_container.get_current_step()
        #
        # if current_step:
        #     current_step.add_info(links)
        #
        # self.test_report.add_info(links)

    def apply_known_bug(self, url: str, description: str = None):
        bug = {
            "url": url,
            "description": description
        }

        current_step = self.step_container.get_current_step()

        if current_step:
            current_step.add_known_bug(bug)

        self.test_report.add_known_bug(bug)

        Integrations.link(bug.get('url'), bug.get('description'))

    def print_message(self, message):
        if isinstance(message, (dict, list)):
            data = json.dumps(message, indent=2, ensure_ascii=False)
            type_ = 'application/json'
        else:
            data = str(message)
            type_ = 'text/plain'

        print_ = Print(data)

        current_step = self.step_container.get_current_step()

        if current_step:
            print_.set_parent_id(current_step.id)
            current_step.add_sub_step(print_)
        else:
            self.step_container.add_step(print_)

        print(f'{data}')

        Integrations.attach(data, print_.id, type_)

    def make_screenshot(self, scr_name: str = None, is_error: bool = False):
        if self.browser is None:
            return

        scr_name = (
            f'{"err_" if is_error else ""}'
            f'scr_{self.test_report.name if not scr_name else scr_name}'
        )

        adapter = get_browser_adapter(self.browser)

        try:
            screenshot_bytes = adapter.make_screenshot()
            if isinstance(screenshot_bytes, list):
                for scr in screenshot_bytes:
                    self.attach(scr, scr_name)
            else:
                self.attach(screenshot_bytes, scr_name)
        except Exception as e:
            print(f'[BLogger][ERROR] Unable to make screenshot: {e}')

    def make_step_err_scr(self, step):
        if self.browser:
            adapter = get_browser_adapter(self.browser)
            screenshot_bytes = adapter.make_screenshot()
            if isinstance(screenshot_bytes, list):
                for scr in screenshot_bytes:
                    step.add_attachment(Attachment(scr))
            else:
                step.add_attachment(Attachment(screenshot_bytes))

    def attach(self, content: Union[bytes, Path, BinaryIO, str, dict, list, int, float, bool, None], name: str = None):

        attachment = Attachment(content=content, name=name)

        current_step = self.step_container.get_current_step()
        if current_step:
            current_step.add_attachment(attachment)

        self.test_report.add_attachment(attachment)

        Integrations.attach(content, attachment.name, attachment.type_)



import os
from glob import glob
from filelock import FileLock

from b_logger.entities.reports import RunReport
from b_logger.utils.paths import b_logs_path, clear_b_logs_tmp, b_logs_tmp_reports_path


class ReportGenerator:
    def __init__(self):
        self.combined = RunReport()

    def generate_combined_report(self):
        self.load_reports()
        self.combined.set_end_time()
        self.combined.count_duration()
        self.save()
        self.clear_locks()

    def load_reports(self):
        report_files = glob(f'{b_logs_tmp_reports_path()}/report_*.json')
        for rep in report_files:
            try:
                report = RunReport.from_json(rep)
                self.merge(report)
            except Exception as e:
                print(f'[BLogger][ERROR] Failed to process {rep}: {e}')
                raise e

    def merge(self, report: RunReport):
        self._merge_proj_name(report)
        self._merge_env(report)
        self._merge_base_url(report)
        self._merge_start_time(report)
        self._merge_end_time(report)
        self._merge_report_ids(report)
        self._merge_run_results(report)
        self._merge_module_results(report)

    def save(self, filename='blog_report'):
        output_path = f'{b_logs_path()}/{filename}'
        with FileLock(f'{output_path}.lock'):
            self.combined.to_json_file(output_path)
        # print(f"[INFO] Combined report saved to {output_path}")

    @staticmethod
    def clear_locks():
        for each in os.listdir(f'{b_logs_path()}'):
            if '.lock' in each:
                os.remove(f'{b_logs_path()}/{each}')

    def _merge_proj_name(self, report: RunReport):
        if self.combined.proj_name and self.combined.proj_name != report.proj_name:
            print(f'[BLogger][WARN] Inconsistent project name: {self.combined.proj_name} vs {report.proj_name}')

        self.combined.proj_name = report.proj_name

    def _merge_env(self, report: RunReport):
        if self.combined.env and self.combined.env != report.env:
            print(f'[BLogger][WARN] Inconsistent env: {self.combined.env} vs {report.env}')
            return

        self.combined.env = report.env

    def _merge_base_url(self, report: RunReport):
        if self.combined.base_url and self.combined.base_url != report.base_url:
            print(f'[BLogger][WARN] Inconsistent base url: {self.combined.base_url} vs {report.base_url}')
            return

        self.combined.base_url = report.base_url

    def _merge_start_time(self, report: RunReport):
        if report.start_time and (
            self.combined.start_time is None or report.get_iso_start_time() < self.combined.get_iso_start_time()
        ):
            self.combined.start_time = report.start_time

    def _merge_end_time(self, report: RunReport):
        if report.end_time and (
            self.combined.end_time is None or report.get_iso_end_time() > self.combined.get_iso_end_time()
        ):
            self.combined.end_time = report.end_time

    def _merge_report_ids(self, report: RunReport):
        self.combined.report_ids[report.worker] = report.report_id

    def _merge_run_results(self, report: RunReport):
        for status, count in report.run_results.items():
            self.combined.run_results.increase(status, count)
            # self.combined.run_results[status] += count

    def _merge_module_results(self, report: RunReport):
        self.combined.combine_modules_from_report(report)



import json
import mimetypes
from contextlib import contextmanager, ExitStack
from abc import ABC, abstractmethod
from pprint import pformat


from b_logger.config import blog_config
from b_logger.entities.attachments import Attachment


# try:
#     from qase.pytest import qase
# except Exception as e:
#     qase = None
#     blog_config.qase = False
#     print(f'[WARN] Qase import error: {e}')
#
# try:
#     import allure
#     from allure_commons.types import AttachmentType
# except Exception as e:
#     allure = None
#     AttachmentType = None
#     blog_config.allure = False
#     print(f'[WARN] Allure import error: {e}')


class IntegrationBase(ABC):
    @abstractmethod
    def is_enabled(self) -> bool: ...

    @abstractmethod
    @contextmanager
    def step(self, title, expected=None): ...

    @abstractmethod
    def description(self, text): ...

    @abstractmethod
    def info(self, name, value): ...

    @abstractmethod
    def link(self, url, name): ...

    @abstractmethod
    def attach(self, content, name, type_): ...


class AllureAdapter(IntegrationBase):
    def __init__(self):
        try:
            import allure
            from allure_commons.types import AttachmentType
            self._allure = allure
            self._AttachmentType = AttachmentType
        except ImportError:
            self._allure = None

    def is_enabled(self):
        return bool(blog_config.allure) and self._allure is not None

    @contextmanager
    def step(self, title, expected=None):
        if self._allure:
            with self._allure.step(title):
                yield
        else:
            yield

    def description(self, text):
        if self._allure:
            self._allure.dynamic.description(text)

    def info(self, name, value):
        if not self._allure:
            return

        try:
            # self._allure.dynamic.parameter(name, value)
            self._allure.attach(str(value), name, self._AttachmentType.TEXT)
        except Exception:
            try:
                self._allure.dynamic.parameter(name, value)
                # self._allure.attach(str(value), name, self._AttachmentType.TEXT)
            except Exception:
                print(f'{name}: {value}')

    def link(self, url, name):
        if self._allure:
            self._allure.dynamic.link(url, name=name)

    def attach(self, content, name, type_):
        if not self._allure:
            return

        mime_type = mimetypes.guess_type(type_)[0] \
            if not type_.startswith("image/") \
            else type_

        mime_map = {
            "image/png": self._AttachmentType.PNG,
            "image/jpeg": self._AttachmentType.JPG,
            "application/json": self._AttachmentType.JSON,
            "text/plain": self._AttachmentType.TEXT,
            "text/html": self._AttachmentType.HTML,
            "application/xml": self._AttachmentType.XML,
        }

        attach_type = mime_map.get(mime_type, self._AttachmentType.TEXT)
        self._allure.attach(content, name, attach_type)


class QaseAdapter(IntegrationBase):
    def __init__(self):
        try:
            from qase.pytest import qase
            self._qase = qase
        except ImportError:
            self._qase = None

    def is_enabled(self):
        return bool(blog_config.qase) and self._qase is not None

    @contextmanager
    def step(self, title, expected=None):
        if self._qase:
            with self._qase.step(title, expected):
                yield
        else:
            yield

    def description(self, text):
        if self._qase:
            self._qase.description(text)

    def info(self, name, value):
        if self._qase:
            if isinstance(value, (dict, list, tuple, set)):
                content = json.dumps(value, indent=2, ensure_ascii=False)
                mimetype = 'application/json'
            elif isinstance(value, (str, int, float, bool)):
                content = str(value)
                mimetype = 'text/plain'
            elif isinstance(value, bytes):
                content = value
                mimetype = 'application/octet-stream'
            else:
                content = pformat(value)
                mimetype = 'text/plain'

            self._qase.attach((content, name, mimetype))

    def link(self, url, name):
        pass

    def attach(self, content, name, type_):
        if self._qase:
            self._qase.attach((content, name, type_))


adapters = [QaseAdapter(), AllureAdapter()]


class Integrations:
    @staticmethod
    def enabled():
        return [a for a in adapters if a.is_enabled()]

    @staticmethod
    @contextmanager
    def step(title, expected=None):
        with ExitStack() as stack:
            for adapter in adapters:
                if adapter.is_enabled():
                    stack.enter_context(adapter.step(title, expected))
            yield

    @staticmethod
    def description(text):
        for a in adapters:
            if a.is_enabled():
                a.description(text)

    @staticmethod
    def info(name, value):
        for a in adapters:
            if a.is_enabled():
                a.info(name, value)

    @staticmethod
    def link(url, name):
        for a in adapters:
            if a.is_enabled():
                a.link(url, name)

    @staticmethod
    def attach(content, name, type_):
        for a in adapters:
            if a.is_enabled():
                a.attach(content, name, type_)

#
# try:
#     from qase.pytest import qase
# except Exception as e:
#     qase = None
#     blog_config.qase = False
#     print(f'[WARN] Qase import error: {e}')
#
# try:
#     import allure
#     from allure_commons.types import AttachmentType
# except Exception as e:
#     allure = None
#     AttachmentType = None
#     blog_config.allure = False
#     print(f'[WARN] Allure import error: {e}')


# class Integrations:
#     qase_enabled: bool = blog_config.qase
#     allure_enabled: bool = blog_config.allure
#
#     @staticmethod
#     @contextmanager
#     def step(title, expected):
#         if Integrations.qase_enabled and Integrations.allure_enabled:
#             with qase.step(title, expected), allure.step(title):
#                 yield
#         elif Integrations.qase_enabled:
#             with qase.step(title, expected):
#                 yield
#         elif Integrations.allure_enabled:
#             with allure.step(title):
#                 yield
#         else:
#             yield
#
#     @staticmethod
#     def description(description):
#         if Integrations.qase_enabled:
#             qase.description(description)
#         if Integrations.allure_enabled:
#             allure.dynamic.description(description)
#
#     @staticmethod
#     def info(name, value):
#         if Integrations.qase_enabled:
#             qase.fields((name, value))
#         if Integrations.allure_enabled:
#             allure.dynamic.parameter(name, value)
#
#     @staticmethod
#     def link(url):
#         if Integrations.qase_enabled:
#             qase.id(url)
#
#         if Integrations.allure_enabled:
#             allure.dynamic.link(url)
#
#
#     @staticmethod
#     def attach(source, attachment: Attachment):
#         def _guess_allure_type(filename_or_mime: str) -> AttachmentType:
#             mime_type = mimetypes.guess_type(filename_or_mime)[0] \
#                 if not filename_or_mime.startswith("image/") \
#                 else filename_or_mime
#
#             return {
#                 "image/png": AttachmentType.PNG,
#                 "image/jpeg": AttachmentType.JPG,
#                 "application/json": AttachmentType.JSON,
#                 "text/plain": AttachmentType.TEXT,
#                 "text/html": AttachmentType.HTML,
#                 "application/xml": AttachmentType.XML,
#             }.get(mime_type, AttachmentType.TEXT)
#
#         if Integrations.qase_enabled:
#             qase.attach((source, attachment.name, attachment.type_))
#
#         if Integrations.allure_enabled:
#             allure_type = _guess_allure_type(attachment.type_)
#             allure.attach(source, attachment.name, allure_type)


# def apply_decorators(*decorators):
#     def meta_decorator(func):
#         for decorator in reversed(decorators):
#             func = decorator(func)
#         return func
#     return meta_decorator



import pytest
import random
from pytest_playwright.pytest_playwright import Playwright
from selenium import webdriver

from b_logger import blog

blog.set_env('prod')

blog.set_base_url('https://base-url.url')


@pytest.fixture()
def some_fixture():
    with blog.step('aaa'):
        pass

    yield

    with blog.step('bbb'):
        pass


@blog.description(
    'Test with base functionality, '
    'this description can be modified inside the test'
)
@blog.info(
    info_explanation='You can leave any useful information by using blog.info()',
    meta={'platform': 'linux', 'python_version': 3.12}
)
@blog.link(
    first_link='http://aaa.com',
    second_link='http://bbb.com'
)
@blog.known_bug(
    'https://link-to-your-bug/1.com',
    'Known Bug Description or Name'
)
def test_main_functionality(some_fixture):
    blog.description('This description will also be added')

    with blog.step('Step 1', 'Step is expected to pass'):
        data = {"a": 1, "b": 2}
        blog.print(f'Some data: {data}')

        blog.link(third_link='http://ccc.com')

        with blog.step('Step 1.1'):
            step_param_1 = random.randint(1, 100)
            step_param_2 = random.randint(1, 100)

            blog.info(
                step_param_1=step_param_1,
                step_param_2=step_param_2
            )

    with blog.step('Step 2'):
        blog.known_bug('https://link-to-your-bug/2.com',
                       # 'Known Bug for a step'
                       )

        with blog.step('Step 2.1'):
            pass

            with blog.step('Step 2.11'):
                pass

                with blog.step('Step 2.111'):
                    pass

                    with blog.step('Step 2.1111'):
                        pass


@pytest.mark.parametrize('py_param_1, py_param_2', [(111, 222), (333, 444)])  # <-- These parameters will be added to test automatically
def test_parametrized(py_param_1, py_param_2):
    with blog.step('Step 1'):
        with blog.step('Step 2'):
            with blog.step('Step 3', 'Step is expected to fail'):
                blog.print(py_param_1)
                with blog.step('Step 4'):
                    blog.print(py_param_2)
                assert py_param_1 in [111, 444]


@pytest.fixture()
def selenium_driver():
    chrome_options = webdriver.ChromeOptions()
    chrome_options.add_argument('--headless=new')

    driver = webdriver.Chrome(options=chrome_options)

    driver.set_window_size(1920, 1080)

    # blog.set_browser(driver) can be also added here, which is preferred

    yield driver

    driver.quit()


@blog.description('This test will make browser screenshot as we did blog.set_browser. '
                  'We can also do it in "selenium_driver" fixture')
@blog.info(run_requirement='To run this test you\'ll need to download chromedriver and put it in your python folder')
def test_selenium_with_set_browser(selenium_driver):
    blog.set_browser(selenium_driver)

    with blog.step('Open any URL'):
        selenium_driver.get(f'https://google.com')

        with blog.step('Raise fake error to check error screenshot'):
            assert 1 == 2


@blog.description('This test will also make browser screenshot as it '
                  'found driver automatically '
                  'based on the following possible browser instance fixture names: '
                  '["driver", "page", "selenium_driver", "driver_init", "playwright_page"]')
@pytest.mark.smoke('weed', every='day')
def test_selenium_without_set_browser(selenium_driver):  #  <-- Will be detected automatically
    blog.info(run_requirement='To run this test you\'ll need to download chromedriver and put it in your python folder')

    with blog.step('Open any URL'):
        selenium_driver.get(f'https://google.com')

        with blog.step('Raise fake error to check error screenshot'):
            print(empty_variable)


@pytest.fixture()
def playwright_page(playwright: Playwright):
    browser = playwright.chromium.launch(headless=True)

    context = browser.new_context()

    page = context.new_page()
    page.set_viewport_size({"width": 1920, "height": 1080})

    # blog.set_browser(page) can also be added here, which is preferred

    yield page

    browser.close()


@pytest.mark.xfail
def test_playwright(playwright_page):  #  <-- Will be detected automatically
    with blog.step('Open any URL'):
        playwright_page.goto(f'https://google.com')

        c2 = playwright_page.context.browser.new_context()
        p2 = c2.new_page()
        p2.bring_to_front()
        p2.goto(f'https://en.wikipedia.org/wiki/Jalape%C3%B1o')

        with blog.step('Raise fake error to check error screenshot'):
            print(empty_variable)



from typing import Optional
import yaml
from pathlib import Path

from b_logger.utils.paths import pathfinder


class BLoggerConfig:
    _instance: Optional["BLoggerConfig"] = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    @classmethod
    def get(cls) -> "BLoggerConfig":
        if cls._instance is None:
            cls._instance = BLoggerConfig()
        return cls._instance

    def __init__(self, path: str = f'{pathfinder.project_root()}/blog.config.yaml'):
        self._extra = {}

        config_path = Path(path)
        self._data = self._load_config_file(config_path)

        # YAML
        self.project_name: Optional[str] = self._data.get("project_name")
        self.env: Optional[str] = self._data.get("env")
        self.base_url: Optional[str] = self._data.get("base_url")

        self.integrations = self._data.get("integrations", {})
        self.qase: bool = bool(self.integrations.get("qase", False))
        self.allure: bool = bool(self.integrations.get("allure", False))

        self.links = self._data.get("links", {})

        # Other
        self._extra = {
            k: v for k, v in self._data.items()
            if not hasattr(self, k)
        }

    @staticmethod
    def _load_config_file(path: Path = None) -> dict:
        if not path.exists():
            raise FileNotFoundError(f'[BLogger][ERROR] blog.config.yaml file not found: {path}')
        with path.open("r", encoding="utf-8") as f:
            return yaml.safe_load(f) or {}

    def apply_cli_options(self, config):
        for opt_name in config.option.__dict__:
            if opt_name.startswith("blog_"):
                value = getattr(config.option, opt_name)
                if value is not None:
                    field_name = opt_name.replace("blog_", "")
                    setattr(self, field_name, value)

    def __getitem__(self, key: str):
        return getattr(self, key, self._extra.get(key, None))

    def __setitem__(self, key: str, value):
        if hasattr(self, key):
            setattr(self, key, value)
        else:
            self._extra[key] = value

    def __getattr__(self, key):
        _extra = object.__getattribute__(self, "_extra") if "_extra" in self.__dict__ else {}
        if key in _extra:
            return _extra[key]
        print(f'[BLogger][WARN] blog_config object has no attribute "{key}"')

    def __setattr__(self, key, value):
        if key in {
            "_data", "_extra", "project_name", "env", "base_url",
            "qase", "allure", "links"
        }:
            object.__setattr__(self, key, value)
        else:
            self._extra[key] = value

    def as_dict(self) -> dict:
        base = {
            "project_name": self.project_name,
            "env": self.env,
            "base_url": self.base_url,
            "qase": self.qase,
            "allure": self.allure,
            "links": self.links,
        }
        return {**base, **self._extra}

    def __repr__(self):
        return f"<BLoggerConfig {self.as_dict()}>"


blog_config = BLoggerConfig.get()



import json
from enum import Enum
from uuid import UUID
from datetime import datetime, timedelta
from collections import defaultdict
from pathlib import Path
from typing import Any


class BaseDataModel:
    indent = 4
    sort_keys = False

    def __str__(self) -> str:
        return self.to_json()

    @staticmethod
    def custom_serializer(obj: Any):
        if isinstance(obj, BaseDataModel):
            return obj.to_dict()
        elif isinstance(obj, Enum):
            return obj.value
        elif isinstance(obj, (UUID, datetime, timedelta, Path)):
            return str(obj)
        elif isinstance(obj, set):
            return list(obj) if len(obj) != 0 else None
        elif isinstance(obj, defaultdict):
            return dict(obj)
        elif isinstance(obj, dict):
            return {k: BaseDataModel.custom_serializer(v) for k, v in obj.items()} if len(obj) != 0 else None
        elif isinstance(obj, list):
            return [BaseDataModel.custom_serializer(v) for v in obj] if len(obj) != 0 else None
        elif hasattr(obj, '__dict__'):
            return vars(obj)
        return str(obj)

    def to_json(self) -> str:
        return json.dumps(self,
                          default=self.custom_serializer,
                          indent=self.indent,
                          sort_keys=self.sort_keys
                          )

    def to_json_file(self, path: str):
        full_path = f'{path}.json'
        with open(full_path, "w", encoding="utf-8") as file:
            json.dump(self,
                      file,
                      default=self.custom_serializer,
                      indent=self.indent,
                      sort_keys=self.sort_keys
                      )

    def to_dict(self) -> dict:
        def convert(value: Any):
            if isinstance(value, Enum):
                return value.value
            if isinstance(value, (UUID, datetime, timedelta, Path)):
                return str(value)
            if isinstance(value, defaultdict):
                return {k: convert(v) for k, v in value.items()} if len(value) != 0 else None
            if isinstance(value, dict):
                return {k: convert(v) for k, v in value.items()} if len(value) != 0 else None
            if isinstance(value, list):
                return [convert(v) for v in value] if len(value) != 0 else None
            if isinstance(value, set):
                return list(value) if len(value) != 0 else None
            return value

        return {k: convert(v) for k, v in self.__dict__.items()}

    @classmethod
    def from_json(cls, filepath: str):
        with open(filepath, "r", encoding="utf-8") as file:
            data = json.load(file)
        return cls.from_dict(data)

    @classmethod
    def from_dict(cls, data: dict):
        if isinstance(data, list):
            return data

        obj = cls()

        for key, value in data.items():
            if isinstance(obj, dict):
                obj[key] = value
            elif hasattr(obj, key):
                setattr(obj, key, value)
            else:
                obj.__dict__[key] = value

        return obj



from abc import ABC, abstractmethod
import importlib


class BrowserAdapter(ABC):
    @abstractmethod
    def make_screenshot(self) -> bytes:
        ...


class SeleniumAdapter(BrowserAdapter):
    def __init__(self, driver):
        self.driver = driver

    def make_screenshot(self) -> bytes:
        return self.driver.get_screenshot_as_png()

    @classmethod
    def is_valid(cls, obj) -> bool:
        try:
            selenium_module = importlib.import_module("selenium.webdriver.ie.webdriver")
            return isinstance(obj, (selenium_module.RemoteWebDriver, selenium_module.WebDriver))
        except ImportError:
            return False


class PlaywrightAdapter(BrowserAdapter):
    def __init__(self, page):
        self.page = page

    def make_screenshot(self) -> bytes | list | None:
        def try_screenshot(page):
            try:
                if page.is_closed():
                    return None
                return page.screenshot(animations='disabled')
            except Exception as e:
                print(f'[BLogger][WARN] Screenshot failed on page: {getattr(page, 'url', '?')}: {e}')
                return None

        # scr = try_screenshot(self.page)
        # if scr:
        #     return scr
        #
        # try:
        #     for context in self.page.context.browser.contexts:
        #         for other_page in context.pages:
        #             if other_page == self.page:
        #                 continue
        #             scr = try_screenshot(other_page)
        #             if scr:
        #                 return scr
        # except Exception as e:
        #     print(f'[BLogger][WARN] Fallback screenshot iteration failed: {e}')
        #
        # print('[BLogger][WARN] All playwright screenshot attempts failed')
        # return None

        scr_container = []

        try:
            for context in self.page.context.browser.contexts:
                for page in context.pages:
                    scr = try_screenshot(page)
                    if scr:
                        scr_container.append(scr)

            return scr_container[0] if len(scr_container) == 1 else scr_container
        except Exception as e:
            print(f'[BLogger][WARN] Screenshot failed: {e}')

    @classmethod
    def is_valid(cls, obj) -> bool:
        try:
            page = importlib.import_module("playwright.sync_api").Page
            return isinstance(obj, page)
        except ImportError:
            return False


ADAPTERS = [SeleniumAdapter, PlaywrightAdapter]


def get_browser_adapter(browser) -> BrowserAdapter:
    for adapter_cls in ADAPTERS:
        if adapter_cls.is_valid(browser):
            return adapter_cls(browser)
    raise RuntimeError(f"Unsupported browser object: {browser}")



import pytest

from pathlib import Path
from typing import Union, BinaryIO
from contextlib import contextmanager

from b_logger.entities.steps import Step
from b_logger.entities.exceptions import possible_exceptions
from b_logger.integrations import Integrations
from b_logger.plugin import runtime


class BLogger:

    @staticmethod
    def set_base_url(base_url: str):
        """
        Set base_url for the entire Run

        Can also be added in blog.config.yaml:
            base_url: 'https://base-url.com'

        Or via command line options:
            --blog_base_url 'https://base-url.com'
        """
        runtime.set_base_url(base_url)

    @staticmethod
    def set_env(env: str):
        """
        Set env for the entire Run

        Can also be added in blog.config.yaml:
            env: 'prod'

        Or via command line options:
            --blog_env 'prod'
        """
        runtime.set_env(env)

    @staticmethod
    def set_browser(browser: "RemoteWebDriver | WebDriver | Page"):
        """
        Set browser in a browser init fixture or in a test

        If browser init fixture name is in
            ["driver", "page", "selenium_driver", "driver_init", "playwright_page"]
        then it will be detected automatically

        Usage:
            @pytest.fixture()
            def selenium_driver():
                driver = webdriver.Chrome()

                blog.set_browser(driver)

                yield driver

                driver.quit()

            or

            def test_playwright(page): <-- Will be detected automatically
                page.goto(f'https://google.com')
        """
        runtime.set_browser(browser)

    @staticmethod
    def description(description: str):
        """
        Add Test Description
        Will be added to the Overview Tab

        Can be used as marker @blog.description() as well as function blog.description()
        Usage inside a test expands description inside marker

        Usage:
            @blog.description(
                'Test with base functionality, '
                'this description can be modified inside the test'
            )
            def test_main_functionality():
                blog.description('This description will also be added')
        """
        runtime.apply_description(description)

        return pytest.mark.blog_description(description=description)

    @staticmethod
    def info(**kwargs):
        """
        Leave any important info or note about Test or Step
        Will be added to the Overview Tab and Current Step

        Can be used as marker @blog.info(k=v) as well as function blog.info(k=v)
            k is a name of an info block
            v supports any data type, but {} is most readable and convenient

        Any amount of info blocks is allowed: @blog.info(k=v, k=v, k=v, ...)

        Usage:
            @blog.info(                             <-- Will be added for a Test
                meta={'platform': 'linux', 'python_version': 3.12}
                additional_parameters=['param 1', 'param 2'],
                some_info='some info',
            )
            def test_main_functionality():
                blog.info(a='a')                    <-- Will be added for a Test

                with blog.step('Step 1'):
                    blog.info(                      <-- Will be added for a Test and a Step
                        step_1_info={'b': 2, 'c': 3}
                    )
        """
        runtime.apply_info(**kwargs)

        return pytest.mark.blog_info(kwargs=kwargs)

    @staticmethod
    def link(**kwargs):
        """
        Attach links to Test or Step
        Will be added to the Overview Tab and Current Step

        Can be used as marker @blog.link(k=v) as well as function blog.link(k=v)
            k is a name of a link
            v is a URL

        Any amount of links is allowed: @blog.link(k=v, k=v, k=v, ...)

        Usage:
            @blog.link(
                first_link='http://aaa.com',
                second_link='http://bbb.com'
            )
            def test_main_functionality():
                blog.link(third_link='http://ccc.com')
        """
        runtime.apply_link(**kwargs)

        return pytest.mark.blog_link(kwargs=kwargs)

    @staticmethod
    def known_bug(url: str, description: str = None):
        """
        Add known bug for Test or Step
        Will be added to the Overview Tab and Current Step

        Usage:
            @blog.known_bug(                            <-- Will be added for a Test
                'Test Bug 1',
                'https://link-to-your-bug/1.com'
            )
            def test_main_functionality():
                blog.known_bug(                         <-- Will be added for a Test
                    'Test Bug 2',
                    'https://link-to-your-bug/2.com'
                )

                with blog.step('Step Title'):           <-- Will be added for a Test and a Step
                    blog.known_bug(
                        'Step Bug',
                        'https://link-to-your-bug/3.com'
                    )
        """
        runtime.apply_known_bug(url, description)

        return pytest.mark.blog_known_bug(url=url, description=description)

    @staticmethod
    @contextmanager
    def step(title: str, expected: str = None):
        """
        Usage:
            with blog.step('Step Title', 'Expected Result'):
                ...
        """
        with Integrations.step(title, expected):

            step = Step(title=title, expected=expected)

            runtime.start_step(step)

            try:
                yield
                runtime.handle_step_result(step)

            except possible_exceptions as e:
                runtime.handle_step_result(step, e)
                raise e

            finally:
                runtime.finish_step(step)

    @staticmethod
    def print(data):
        """
        Print any message (str, dict, list, object, etc.)
        Will be added to a Current Step

        Usage:
            data = {"a": 1, "b": 2}
            blog.print(f'Some important data: {data}')

            blog.print(f'Probably too long str\n'
                        'can be newlined like that')
        """
        runtime.print_message(data)

    @staticmethod
    def screenshot(name: str = None, is_error: bool = False):
        """
        Make screenshot
        Will be added to the Attachments Tab and Current Step

        Will do nothing if no browser is used

        Usage:
            blog.screenshot('scr_name')
            blog.screenshot('err_scr_name', True)
        """
        runtime.make_screenshot(name, is_error)

    @staticmethod
    def attach(content: Union[bytes, Path, BinaryIO, str, dict, list, int, float, bool, None], name: str = None):
        """
        Attach file or data
        Will be added to the Attachments Tab and Current Step

        Usage:
            blog.attach({"a": 1, "b": 2}, 'some_data')
        """
        runtime.attach(content, name)
